package fa;

import fa.util.*;

import ec.*;
import ec.util.*;

public class AlgebraStatistics extends GPBooleanMultiStatistics
{

public Individual[] best_malcev_of_run, best_pixley_of_run;

private Individual[] best_malcev_of_gen, best_pixley_of_gen;

/*
 * Is Individual "a" smaller than Individual "b"?
 * Fitness as a discriminator breaks ties; tree height breaks further ties.
 */
public boolean isSmaller(final Individual a, final Individual b)
{                 
	return (gpSize(a) < gpSize(b) ||
		(gpSize(a) == gpSize(b) &&
		 (a.fitness.betterThan(b.fitness) ||
		  (a.fitness.equivalentTo(b.fitness) && gpHeight(a) < gpHeight(b)))));
}

/*
 * Is Individual "a" better than Individual "b"?
 * Tree size breaks ties; fitness as a Mal'cev term breaks further ties;
 *  fitness as a Pixley term breaks still further ties; any remaining
 *  ties can potentially be broken by tree height.
 */
public boolean isBetter(final Individual a, final Individual b)
{
	final AlgebraFitness af = (AlgebraFitness)a.fitness;
	final AlgebraFitness bf = (AlgebraFitness)b.fitness;

	return (af.misses < bf.misses ||
		(af.misses == bf.misses &&
		 (gpSize(a) < gpSize(b) ||
		  (gpSize(a) == gpSize(b) &&
		   (af.malcev < bf.malcev ||
		    (af.malcev == bf.malcev &&
		     (af.pixley < bf.pixley ||
		      (af.pixley == bf.pixley && gpHeight(a) < gpHeight(b)))))))));
}

private void perIndividual(final Individual ind, final int sp)
{
	AlgebraFitness f = (AlgebraFitness)ind.fitness;

	if (f.malcev == 0) {
		if (best_malcev_of_gen[sp] == null ||
		    isSmaller(ind, best_malcev_of_gen[sp]))
			best_malcev_of_gen[sp] = ind;
		if (f.pixley == 0 &&
		    (best_pixley_of_gen[sp] == null ||
		     isSmaller(ind, best_pixley_of_gen[sp])))
			best_pixley_of_gen[sp] = ind;
	}
}

private void perGeneration(final EvolutionState state)
{
	for (int i = 0; i < best_malcev_of_gen.length; ++i) {
		final Individual pix = best_pixley_of_gen[i];
		if (pix != null) {
			if (best_pixley_of_run[i] == null ||
			    isSmaller(pix, best_pixley_of_run[i]))
				best_pixley_of_run[i] = (Individual)pix.clone();
			if (pix != best_of_gen[i]) {
				state.output.println("Best Pixley term:\n" +
						     sizeToString(pix),
						     Output.V_NO_GENERAL,
						     statisticslog);
				pix.printIndividualForHumans(state,
							     statisticslog,
							     Output.V_NO_GENERAL);
			}
			best_pixley_of_gen[i] = null;
		}
		final Individual mal = best_malcev_of_gen[i];
		if (mal != null) {
			if (best_malcev_of_run[i] == null ||
			    isSmaller(mal, best_malcev_of_run[i]))
				best_malcev_of_run[i] = (Individual)mal.clone();
			if (mal != pix && mal != best_of_gen[i]) {
				state.output.println("Best Mal'cev term:\n" +
						     sizeToString(mal),
						     Output.V_NO_GENERAL,
						     statisticslog);
				mal.printIndividualForHumans(state,
							     statisticslog,
							     Output.V_NO_GENERAL);
			}
			best_malcev_of_gen[i] = null;
		}
	}
}

private void perRun(final EvolutionState state)
{
	final int n = state.population.subpops.length;

	best_malcev_of_run = new Individual[n];
	best_malcev_of_gen = new Individual[n];
	best_pixley_of_run = new Individual[n];
	best_pixley_of_gen = new Individual[n];
}

public void finalStatistics(final EvolutionState state, final int result)
{
	super.finalStatistics(state, result);

/*
	for (int i = 0; i < best_pixley_of_run.length; ++i) {
		final Individual pix = best_pixley_of_run[i];
		if (pix != best_of_run[i]) {
			state.output.println("\nBest Pixley term of run:", 
					     Output.V_NO_GENERAL,
					     statisticslog);
			pix.printIndividualForHumans(state, statisticslog,
						     Output.V_NO_GENERAL);
		}
		final Individual mal = best_malcev_of_run[i];
		if (mal != pix && mal != best_of_run[i]) {
			state.output.println("\nBest Mal'cev term of run:", 
					     Output.V_NO_GENERAL,
					     statisticslog);
			mal.printIndividualForHumans(state, statisticslog,
						     Output.V_NO_GENERAL);
		}
	}
*/
}

}
