package fa.util;

import java.util.*;

import ec.*;
import ec.util.*;

public class AlphaBreeder extends Breeder implements DynamicTourneySize,
						     BreedingIndex
{
	public class Worker extends Thread
	{
		public int thread;
		public BreedChunk c;

		public Worker(int thread)
		{
			this.thread = thread;
		}

		public void run()
		{
			while (true) {
				synchronized (AlphaBreeder.this) {
					while (workq.isEmpty()) {
						try {
							AlphaBreeder.this.wait();
						} catch (InterruptedException e) {}
					}
					++working;
					c = workq.remove();
				}
				c.run(thread);
				c = null;
				synchronized (AlphaBreeder.this) {
					--working;
					if (working == 0)
						AlphaBreeder.this.notifyAll();
				}
			}
		}
	}

	public static class BreedChunk
	{
		public EvolutionState state;
		public Individual[] inds;
		public int sp, min, max;

		public int ind;

		public BreedChunk(EvolutionState state, int sp, Individual[] inds,
				  int min, int max)
		{
			this.state = state;
			this.sp = sp;
			this.inds = inds;
			this.min = min;
			this.max = (max < inds.length) ? max : inds.length;
		}

		public void run(int thread)
		{
			BreedingPipeline bp = (BreedingPipeline)state.population.subpops[sp].species.pipe_prototype.clone();
			bp.prepareToProduce(state, sp, thread);

			for (ind = min; ind < max; ++ind)
				bp.produce(1, 1, ind, sp, inds, state, thread);

			bp.finishProducing(state, sp, thread);
		}
	}

	public double tsize;

	public Queue<BreedChunk> workq = new LinkedList<BreedChunk>();
	public int working = 0;
	public Worker[] threads;

	public Population breedPopulation(EvolutionState state)
	{
		int psize = 0;
		for (Subpopulation sp : state.population.subpops)
			psize += sp.individuals.length;

		int a;
		try {
			a = ((GPBooleanMultiStatistics)state.statistics).getAlpha();
		} catch (ClassCastException e) {
			a = AlphaSet.findAlpha(state.population);
		}
//		tsize = AlphaSet.tourneySize((100.0 * a) / psize);
		tsize = AlphaSet.tourneySize(Math.min(100.0,
						      (100.0 * a) / 5000.0));

		Population newpop = (Population)state.population.emptyClone();

		for (int sp = 0; sp < newpop.subpops.length; ++sp)
			synchronized (this) {
				Individual[] inds = newpop.subpops[sp].individuals;
				int n = inds.length;
				int m = (int)Math.ceil(n / (2.0 * state.breedthreads));

				int j;
				for (int i = 0; i < n; i = j) {
					j = i + m;
					workq.add(new BreedChunk(state, sp, inds,
								 i, j));
				}
				notifyAll();
			}

		if (threads != null)
			synchronized (this) {
				while (!workq.isEmpty())
					try {
						wait();
					} catch (InterruptedException e) {}
				while (working > 0)
					try {
						wait();
					} catch (InterruptedException e) {}
			}
		else
			while (!workq.isEmpty()) {
				workq.peek().run(0);
				workq.remove();
			}

		return newpop;
	}

	public double tourneySize() { return tsize; }

	public int breedingIndex(int thread)
	{
		return ((threads != null) ? threads[thread].c : workq.peek()).ind;
	}

	public void setup(EvolutionState state, Parameter base)
	{
		if (state.breedthreads > 1) {
			threads = new Worker[state.breedthreads];
			for (int i = 0; i < state.breedthreads; ++i) {
				threads[i] = new Worker(i);
				threads[i].setDaemon(true);
				threads[i].start();
			}
		}
	}
}
