package fa.util;

import ec.*;
import ec.util.*;
import java.io.*;

public class BooleanMultiFitnessImpl extends Fitness
				     implements BooleanMultiFitness
{
	public boolean[] p;
	public double f;
	public int misses;

	public void setFitness(boolean[] p)
	{
		this.p = p;

		misses = 0;
		for (boolean c : p)
			if (c)
				++misses;
		f = misses;
	}

	public void scale(double[] s)
	{
		f = 0.0;
		for (int i = 0; i < p.length; ++i)
			if (p[i])
				f += s[i];
	}

	public boolean[] getFitness()
	{
		/*
		 * This really ought to be a const return, but
		 * since Jaba lacks this concept, and I don't
		 * want to copy the whole array...
		 */
		return p;
	}

	public int misses()
	{
		return misses;
	}

	public float fitness()
	{
		return (float)(1.0 / f);
	}

	public boolean isIdealFitness()
	{
		return (misses == 0);
	}

	public boolean equivalentTo(Fitness g)
	{
		return (f == ((BooleanMultiFitnessImpl)g).f);
	}

	public boolean betterThan(Fitness g)
	{
		return (f < ((BooleanMultiFitnessImpl)g).f);
	}

	public String fitnessToStringForHumans()
	{
		return ("Misses: " + misses + "  (scaled: " + f + ")");
	}

	public Object clone()
	{
		BooleanMultiFitnessImpl ret;

		ret = (BooleanMultiFitnessImpl)super.clone();
		ret.p = p;
		ret.f = f;
		ret.misses = misses;

		return ret;
	}

	public Parameter defaultBase()
	{
		return (new Parameter("boolmultifit"));
	}

	public void writeFitness(final EvolutionState state,
				 final DataOutput dataOutput) throws IOException
        {
		dataOutput.writeDouble(f);
		dataOutput.writeInt(misses);
        }

	public void readFitness(final EvolutionState state,
				final DataInput dataInput) throws IOException
        {
		f = dataInput.readDouble();
		misses = dataInput.readInt();
        }
}
