package fa.util;

import ec.*;
import ec.util.*;

public class DynamicSpatialTourneySelection extends SelectionMethod
{
	public double radius, size;

	public void prepareToProduce(EvolutionState state, int sp, int t)
	{
		size = ((DynamicTourneySize)state.breeder).tourneySize();
	}

	public int randomNeighbor(int i, int n, MersenneTwisterFast rng)
	{
		int r = (int)radius;
		return ((i + rng.nextInt(2 * r + 1) - r + n) % n);
	}

	public int produce(int sp, EvolutionState state, int thread)
	{
		MersenneTwisterFast rng = state.random[thread];
		Individual[] oldinds = state.population.subpops[sp].individuals;
		int n = oldinds.length;

		int i = ((BreedingIndex)state.breeder).breedingIndex(thread);
		int j = randomNeighbor(i, n, rng);
		for (int round = 1; round < (int)size; ++round) {
			int k = randomNeighbor(i, n, rng);
			if (oldinds[k].fitness.betterThan(oldinds[j].fitness))
				j = k;
		}
		return j;
	}

	public int produce(final int min, final int max, final int start,
			   final int sp, Individual[] inds,
			   final EvolutionState state, final int thread)
	{
		int i, n;
		Individual[] oldinds = state.population.subpops[sp].individuals;

		n = (max > min) ? max : min;
		if (n < 1)
			n = 1;
		if (n > inds.length)
			n = inds.length;

		for (i = 0; i < n; ++i)
			inds[i + start] = oldinds[produce(sp, state, thread)];
		return i;
	}

	public Parameter defaultBase()
	{
		return (new Parameter("trivialgeo"));
	}

	public void setup(final EvolutionState state, final Parameter base)
	{
		super.setup(state, base);

		radius = state.parameters.getDouble(base.push("radius"),
						    defaultBase().push("radius"),
						    0.0);
		if (radius < 0.0)
			state.output.fatal("Selection radius must be positive.",
					   base.push("radius"),
					   defaultBase().push("radius"));
	}

	public Object clone()
	{
		DynamicSpatialTourneySelection obj = (DynamicSpatialTourneySelection)super.clone();
		obj.radius = radius;
		return obj;
	}
}
