package fa.util;

import java.util.*;

import ec.*;
import ec.gp.*;
import ec.simple.*;
import ec.util.*;

public class GPBooleanMultiStatistics extends SimpleStatistics {

public int[] fit_hist;

private Map<Object, Long> sizes;
private int indSize, indHeight;
private static final long SIZEMASK = 0xffffffffL;
private static final long HEIGHTMASK = 0xffffffff00000000L;

protected Individual[] best_of_gen;

private void perIndividual(final Individual ind, final int sp) {}
private void perGeneration(final EvolutionState state) {}
private void perRun(final EvolutionState state) {}

public void postEvaluationStatistics(final EvolutionState state)
{
	final Population p = state.population;

	if (fit_hist == null) {
		for (int i = 0; i < p.subpops.length; ++i)
			best_of_run[i] = (Individual)p.subpops[i].individuals[0].clone();

		final Individual u = best_of_run[0];
		final BooleanMultiFitness f = (BooleanMultiFitness)u.fitness;
		fit_hist = new int[f.getFitness().length + 1];
		perRun(state);
	} else
		for (int i = 0; i < fit_hist.length; ++i)
			fit_hist[i] = 0;

	best_of_gen = new Individual[p.subpops.length];
	for (int i = 0; i < p.subpops.length; ++i) {
		final Individual[] inds = p.subpops[i].individuals;
		best_of_gen[i] = inds[0];
		for (final Individual ind : inds) {
			++fit_hist[((BooleanMultiFitness)ind.fitness).misses()];
			if (isBetter(ind, best_of_gen[i]))
				best_of_gen[i] = ind;
			perIndividual(ind, i);
		}
	}
	for (int i = 0; i < p.subpops.length; ++i)
		if (isBetter(best_of_gen[i], best_of_run[i]))
			best_of_run[i] = (Individual)best_of_gen[i].clone();

	state.output.println("\nGeneration: " + state.generation +
			     "\nAlpha: " + getAlpha() +
			     "\nBest Individual:", Output.V_NO_GENERAL,
			     statisticslog);
	for (final Individual ind : best_of_gen) {
		state.output.println(sizeToString(ind),
				     Output.V_NO_GENERAL, statisticslog);
		ind.printIndividualForHumans(state, statisticslog,
					     Output.V_NO_GENERAL);
	}
	perGeneration(state);

	best_of_gen = null;
	sizes.clear();
}

public int getAlpha()
{
	for (int i = 0; i < fit_hist.length; ++i)
		if (fit_hist[i] > 0)
			return fit_hist[i];
	assert(false);
	return -1;
}

/*
 * Is Individual "a" better than Individual "b"?
 * Tree size breaks ties; height breaks further ties.
 */
public boolean isBetter(final Individual a, final Individual b)
{
	return (a.fitness.betterThan(b.fitness) ||
		(a.fitness.equivalentTo(b.fitness) &&
		 (gpSize(a) < gpSize(b) ||
		  (gpSize(a) == gpSize(b) && gpHeight(a) < gpHeight(b)))));
}

public int gpSize(final Individual ind)
{
	if (sizes.get((Object)ind) != null)
		return (int)(sizes.get(ind) & SIZEMASK);

	final GPNode root = ((GPIndividual)ind).trees[0].child;

	indSize = indHeight = 0;
	calcSize(root, 0);
	sizes.put((Object)ind, ((long)indHeight << 32) | indSize);

	return indSize;
}

public int gpHeight(final Individual ind)
{
	return (int)((sizes.get((Object)ind) & HEIGHTMASK) >>> 32);
}

private void calcSize(final GPNode p, int d)
{
	++indSize;
	++d;
	if (p.children.length > 0)
		for (final GPNode c : p.children)
			calcSize(c, d);
	else if (d > indHeight)
		indHeight = d;
}

public String sizeToString(final Individual ind)
{
	return "Size: " + gpSize(ind) + "  (height: " + gpHeight(ind) + ")";
}

public void setup(final EvolutionState state, final Parameter base)
{
	super.setup(state, base);

	sizes = new WeakHashMap<Object, Long>();
}

}
