/*
  Copyright 2006 by Sean Luke
  Licensed under the Academic Free License version 3.0
*/

package fa.util;

import ec.*;
import ec.simple.*;
import ec.util.*;

public class HitCountingEvaluator extends Evaluator
{

protected class EvalThread extends Thread
{
	final int[] ni, i0;
	final EvolutionState state;
	final SimpleProblemForm prob;
	final int thr;

	int[][] partial_hits;
	boolean ideal;

	public
	EvalThread(int thr, int[] ni, int[] i0, EvolutionState state,
		   Object prob)
	{
		this.thr = thr;
		this.ni = ni;
		this.i0 = i0;
		this.state = state;
		this.prob = (SimpleProblemForm)prob;

		ideal = false;

		partial_hits = new int[hit_counts.length][];
		for (int i = 0; i < hit_counts.length; ++i)
			partial_hits[i] = new int[hit_counts[i].length];
	}

	public void
	run()
	{
		Subpopulation[] p = state.population.subpops;

		((Problem)prob).prepareToEvaluate(state, thr);
        
		for (int sp = 0; sp < state.population.subpops.length; ++sp)
			for (int i = i0[sp]; i < i0[sp] + ni[sp]; ++i) {
				Individual ind = p[sp].individuals[i];

                		prob.evaluate(state, ind, thr);

				BooleanMultiFitness f = (BooleanMultiFitness)
							    ind.fitness;
				boolean[] m = f.getFitness();

				for (int j = 0; j < m.length; ++j)
					if (!m[j])
						++partial_hits[sp][j];

				f.scale(scale[sp]);

				ideal |= ind.fitness.isIdealFitness();
			}

		((Problem)prob).finishEvaluating(state, thr);
	}
}

	public long[][] hit_counts;

	protected double scale[][];
	protected boolean ideal;

	public void
	evaluatePopulation(final EvolutionState state)
	{
		final Subpopulation[] p = state.population.subpops;
		final int nt = state.evalthreads;

		int[][] ni = new int[nt][p.length];
		int[][] i0 = new int[nt][p.length];

		if (state.generation == 0) {
			SimpleProblemForm prob = (SimpleProblemForm)(p_problem.clone());

			hit_counts = new long[p.length][];
			scale = new double[p.length][];
			for (int sp = 0; sp < p.length; ++sp) {
				Individual ind = p[sp].individuals[0];

				prob.evaluate(state, ind, 0);

				int n = ((BooleanMultiFitness)ind.fitness).getFitness().length;

				hit_counts[sp] = new long[n];
				scale[sp] = new double[n];
				for (int i = 0; i < n; ++i)
					scale[sp][i] = 1.0;
			}
		} else {
			final double g = state.generation;

			for (int sp = 0; sp < p.length; ++sp)
				for (int i = 0; i < scale[sp].length; ++i)
					scale[sp][i] = g / (g +
							    hit_counts[sp][i]);
		}

		ideal = false;

		for (int t = 0; t < nt; ++t)
			for (int sp = 0; sp < p.length; ++sp) {
				int n = p[sp].individuals.length;

				ni[t][sp] = n / nt;
				if (t == nt - 1)
					ni[t][sp] += n - nt * (n / nt);

				i0[t][sp] = t * (n / nt);
			}

		if (nt == 1) {
			EvalThread t = new EvalThread(0, ni[0], i0[0], state,
						      p_problem.clone());
			t.run();
			add_hits(t);
		} else {
			EvalThread[] threads = new EvalThread[nt];

			for (int t = 0; t < nt; ++t) {
				threads[t] = new EvalThread(t, ni[t], i0[t],
							    state,
							    p_problem.clone());
				threads[t].start();
			}

			for (Thread t : threads)
				do
					try {
						t.join();
					} catch (InterruptedException e) {}
				while (t.isAlive());

			for (EvalThread t : threads)
				add_hits(t);
		}
        }

	protected void
	add_hits(EvalThread t)
	{
		ideal |= t.ideal;

		for (int i = 0; i < hit_counts.length; ++i)
			for (int j = 0; j < hit_counts[i].length; ++j)
				hit_counts[i][j] += t.partial_hits[i][j];
	}

	public boolean
	runComplete(final EvolutionState state)
	{
		return ideal;
        }
}
